/*******************************************************************************
*	Packet handler for localizing and anchor node. command.c
*	Last Modified: 21/May/2009
*
*	Copyright (C) 2009 Tony Wu mdzz@u.washington.edu
*
*	This program is free software: you can redistribute it and/or modify
*	it under the terms of the GNU General Public License as published by
*	the Free Software Foundation, either version 3 of the License, or
*	(at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	You should have received a copy of the GNU General Public License
*	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*******************************************************************************/

#include "command.h"
#include "router.h"
#include <pthread.h>

//debug switch
#define cmd_debug 0

// global vars to pass to the GUI
char total_cmds = 0;	// total number of cmds to be sent out by the base node
char sent_cmds = 0;		// number of cmds sent by the base node
__u8 seq_list[MAX_SEQ];	// list of previously received packet sequence number
int list_index = 0;		//
pthread_mutex_t seqList_mutex = PTHREAD_MUTEX_INITIALIZER;	//


/*
 * this function initializes the sequence # list
 */
int init_seq_list()
{
	int i = 0;
	for (i = 0; i < MAX_SEQ; i++)
		seq_list[i] = 0;
}


/*
 * this functin handles all arriving packets
 * packets will be processed with different handler threads
 * inputs:
 *	*in_pkt: incoming packet
 * outputs:
 *	returns 1 if successfully handled, -1 otherwise
 */
int pkt_handler(packet *in_pkt)
{
	int i = 0;
	
	// check if packet is previously received, skipping all broadcast
	if ((in_pkt->type != ANCHOR_BROADCAST) && (in_pkt->type != ROUTING_PKT_TYPE) && (in_pkt->type != TO_NODE_ROUTING_PKT_TYPE)) {
		for (i = 0; i < MAX_SEQ; i++) {
			if (seq_list[i] == in_pkt->seq) {
				printf("\nThis packet has been received before!\n\n");
				return -1;
			}
		}
		pthread_mutex_lock(&seqList_mutex);
		seq_list[list_index] = in_pkt->seq;	// update recorded sequence number
		if (++list_index >= MAX_SEQ)
			list_index = 0;	// reset list index
		pthread_mutex_unlock(&seqList_mutex);
	}
	
	// process different packets
	switch (in_pkt->type) {
		case CMD_PKT_TYPE:
			printf("Command packet.\n");
			forward_to_node(in_pkt, 1);		// ACK on
			break;
		case ANCHOR_BROADCAST:
			printf("Anchor broadcast.\n");
			cmd_handler(in_pkt);
			break;
		case ROUTING_PKT_TYPE:
			printf("Back to base routing packet.\n");
			recv_rtPkt(in_pkt);
			break;
		case TO_NODE_ROUTING_PKT_TYPE:
			printf("Base to nodes routing packet.\n");
			forward_rt_toBase(in_pkt);
			break;
		case TO_BASE_DATA_PKT_TYPE:
			printf("Location data packet.\n");
			forward_to_base(in_pkt, 1);		// ACK on
			break;
		default:
			printf("Undefined packet type.\n");
			return -1;	// undefined command, reutrn error
	}
	return 0;
}


/*
 * this function handles all the command packets
 * inputs:
 *	*in_pkt: incoming packet
 * outputs:
 *	-1 if received command not defined 
 */
int cmd_handler(packet *in_pkt)
{
	printf("Received command # is: %d\n", in_pkt->data[1]);
	switch (in_pkt->data[1]) {
		case BLUE_TOGGLE:
			printf("Toggle blue LED.\n");
			toggle_blue();
			break;
		case BLUE_ON:
			printf("Turn on blue LED.\n");
			on_blue();
			break;
		case BLUE_OFF:
			printf("Turn off blue LED.\n");
			off_blue();
			break;
		case RED_TOGGLE:
			printf("Toggle red LED.\n");
			toggle_red();
			break;
		case RED_ON:
			printf("Turn on red LED.\n");
			on_red();
			break;
		case RED_OFF:
			printf("Turn off red LED.\n");
			off_red();
			break;
		case GREEN_TOGGLE:
			printf("Toggle green LED.\n");
			toggle_green();
			break;
		case GREEN_ON:
			printf("Turn on green LED.\n");
			on_green();
			break;
		case GREEN_OFF:
			printf("Turn off green LED.\n");
			off_green();
			break;
		case ALL_TOGGLE:
			printf("Toggle all LEDs.\n");
			toggle_red();
			toggle_green();
			toggle_blue();
			break;
		case ALL_ON:
			printf("Turn on all LEDs.\n");
			on_red();
			on_green();
			on_blue();
			break;
		case ALL_OFF:
			printf("Turn off all LEDs.\n");
			off_red();
			off_green();
			off_blue();
			break;
		case ASSIGN_XY:
		case LOC_DATA:
		case EST_LOC_REPORT:
		case LOC_OFF:
		case ACTUAL_LOC_REPORT:
		case PROPOGATE_ON:
		case PROPOGATE_OFF:
		case LOC_ON:
		case CENTROID:
		case WEIGHTEDCENTROID:
		case GRIDSCORE:
		case WEIGHTED_SCORE:
		case GRADIENT:
		case WEIGHTED_GRADIENT:
		case CHANGE_MAX_RSSI:
		case CHANGE_LOCAL_TIMER_INTERVAL:
			printf("Handling localization packets.\n");
			localization_msg(in_pkt);
			break;
		case RESTART:
			printf("Restarting system!\n");
			toggle_green();
			sleep(1);
			toggle_green();
			exit(-1);	// restart system by exiting with "-1". the script will capture this and restart the system
		default:
			printf("Undefined command # %d.\n", in_pkt->data[1]);
			return -1;
	}
	return 0;
}

/*
 * this function forwards the msg to base via routing
 */
int forward_to_base(packet *out_pkt, int ACK_en)
{
	__u16 localAddr;	// local address
	FILE *fp;
	
	// read local address form config file
	localAddr = get_addr();
	
	if (check_base()) {
		printf("\nPacket has reached base!\n\n");
		// call cmd handler
		cmd_handler(out_pkt);
		return 0;
	}
	else {
		out_pkt->srcAddr = localAddr;
		if (sendtoBase(out_pkt, ACK_en) < 0)
			return -1;
	}
	return 0;
}

/*
 * this function forwards command packets to nodes 
 */
int forward_to_node(packet *out_pkt, int ACK_en)
{
	__u16 localAddr;	// local address
	__u16 destAddr;	//
	__u16 nextHop;	//
	
	// read local address form config file
	localAddr = get_addr();
	
	// get dest address
	destAddr = out_pkt->data[1];
	destAddr = destAddr << 8;
	destAddr += out_pkt->data[2];
	
	if (localAddr == destAddr) {	// packet arrived at dest
		printf("\nPacket has reached destination node!\n\n");
		// re-construct cmd content
		memcpy(out_pkt->data, out_pkt->data + 14, 12);
		// call cmd handler
		cmd_handler(out_pkt);
		return 0;
	}
	else {
		out_pkt->data[0]--;	// update number of hops left
		out_pkt->srcAddr = localAddr;	//
		// get next hop from the end of the list
		nextHop = out_pkt->data[1 + 2 * (out_pkt->data[0] - 1)];
		nextHop = nextHop << 8;
		nextHop += out_pkt->data[2 + 2 * (out_pkt->data[0] - 1)];
		// send pkt
		printf("Sending packet to %x through: %x\n", destAddr, nextHop);
		if (h_send(nextHop, out_pkt, ACK_en) < 0)
			return -1;
	}
	return 0;
}

/*
 * this function toggles the green LED
 */
int toggle_green()
{
	int gpioControl;
	gpioControl = open(GPIO_CONTROL, O_WRONLY|O_NOCTTY|O_NDELAY);
	write (gpioControl,"r104", 3);
	close (gpioControl);
	int green;
	char buffer[BUF_SIZE];
	green = open(GPIOGREEN, O_RDONLY|O_NOCTTY|O_NDELAY);
	if (cmd_debug)
		printf("green = %d\n", green);
	read(green, buffer, BUF_SIZE-1);
	close(green);
	int number = atoi(buffer);
	if (cmd_debug) {
		printf("number = %d\n", number);
		printf("toggle green LED\n\n");
	}
	if (number == 1)
		system("echo c104 > /proc/gpio/GPCTL");
	else
		system("echo s104 > /proc/gpio/GPCTL");
	return(0);
}

/*
 * this function turns on the green LED
 */
int on_green()
{
	system("echo c104 > /proc/gpio/GPCTL");
	return 0;
}

/*
* this function turns off the green LED
*/
int off_green()
{
	system("echo s104 > /proc/gpio/GPCTL");
	return 0;
}

/*
 * this function toggles the red LED
 */
int toggle_red()
{
	int gpioControl;
	gpioControl = open(GPIO_CONTROL, O_WRONLY|O_NOCTTY|O_NDELAY);
	write (gpioControl,"r103", 3);
	close (gpioControl);
	int red;
	char buffer[BUF_SIZE];
	red = open(GPIORED, O_RDONLY|O_NOCTTY|O_NDELAY);
	if (cmd_debug)
		printf("red = %d\n", red);
	read(red, buffer, BUF_SIZE-1);
	close(red);
	int number = atoi(buffer);
	if (cmd_debug) {
		printf("number = %d\n", number);
		printf("toggle red LED\n\n");
	}
	if (number == 1)
		system("echo c103 > /proc/gpio/GPCTL");
	else
		system("echo s103 > /proc/gpio/GPCTL");
	return(0);
}

/*
* this function turns on the red LED
*/
int on_red()
{
	system("echo c103 > /proc/gpio/GPCTL");
	return 0;
}

/*
* this function turns off the red LED
*/
int off_red()
{
	system("echo s103 > /proc/gpio/GPCTL");
	return 0;
}

/*
 * this function toggles the blue LED
 */
int toggle_blue()
{
	int gpioControl;
	gpioControl = open(GPIO_CONTROL, O_WRONLY|O_NOCTTY|O_NDELAY);
	write (gpioControl,"r105", 3);
	close (gpioControl);
	int blue;
	char buffer[BUF_SIZE];
	blue = open(GPIOBLUE, O_RDONLY|O_NOCTTY|O_NDELAY);
	if (cmd_debug)
		printf("blue = %d\n", blue);
	read(blue, buffer, BUF_SIZE-1);
	close(blue);
	int number = atoi(buffer);
	if (cmd_debug) {
		printf("number = %d\n", number);
		printf("toggle blue LED\n\n");
	}
	if (number == 1)
		system("echo c105 > /proc/gpio/GPCTL");
	else
		system("echo s105 > /proc/gpio/GPCTL");
	return(0);
}

/*
* this function turns on the blue LED
*/
int on_blue()
{
	system("echo c105 > /proc/gpio/GPCTL");
	return 0;
}

/*
* this function turns off the blue LED
*/
int off_blue()
{
	system("echo s105 > /proc/gpio/GPCTL");
	return 0;
}


/*
 * this function sends commands to different destination nodes
 */
/*
int send_cmd(unsigned char *cmd_list)
{
	packet out_pkt;	// out going cmd packet
	int i;
	unsigned char *tmp = cmd_list + 1;	//
	__u16 destAddr;
	
	// initialize pkt
	if (initPkt(&out_pkt, CMD_PKT_TYPE) < 0)
		return -1;
	
	if (cmd_debug)
		printf("Number of commands to be sent is: %d\n", total_cmds);
	
	// send all command pkts 
	for (i = 0; i < total_cmds; i++) {
		sent_cmds = i;				// update number of cmds left
		destAddr = tmp[0];			// get destAddr hi byte
		destAddr = destAddr << 8;	// shift to hi byte
		destAddr += tmp[1];			// get destAddr lo byte
		memcpy(out_pkt.data, tmp + 2, 12);	// get cmd content
		tmp += 14;					// move to next cmd
		if (cmd_debug) {
			printf("Command should be sent to %x\n", destAddr);
			printf("Command is %d\n", out_pkt.data[1]);
		}
		printf("\nSending packet #%d, %d left to send.\n\n", i, total_cmds - i);
		// send pkt via routing
		if (sendtoNode(destAddr, &out_pkt, 1) < 0) {
			printf("Error sending cmd to Node: %x\n", destAddr);
			// report error packet back to GUI
			errorSocket(out_pkt);
		}
		// wait before sending next command
		usleep(WAIT_TIME);	// 1s is the minimum val
	}
	// reset counter
	sent_cmds = 0;
	total_cmds = 0;
	printf("All commands sent!!!\n\n");
	return 0;
}
*/

