/*******************************************************************************
*	Packet sending/receiving. network.c
*	Last Modified: 21/May/2009
*
*	Copyright (C) 2009 Tony Wu mdzz@u.washington.edu
*
*	This program is free software: you can redistribute it and/or modify
*	it under the terms of the GNU General Public License as published by
*	the Free Software Foundation, either version 3 of the License, or
*	(at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	You should have received a copy of the GNU General Public License
*	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*******************************************************************************/

#include "network.h"

// some global vars
pthread_mutex_t radio_mutex = PTHREAD_MUTEX_INITIALIZER;	// flag mutex


/*
 * this function initializes the radio (cc2420) for later communication.
 * input:
 * 	srcAddr: local address
 * output:
 *	return 0 if init success, else -1
 */
int cc2420_init(int pwr_level)
{
	int ret, fd_tosmac;

	// open as blocking mode
	fd_tosmac = open(TOSMAC_DEVICE, O_RDWR);
	if (fd_tosmac < 0)
	{
		fprintf(stderr, "Open error: %s\n", TOSMAC_DEVICE);
		return -1;
	}

	// set freq 2405 to 2480
	if(ioctl(fd_tosmac, TOSMAC_IOSETFREQ, 2405) < 0)
	{
		fprintf (stderr, "IOCTL set TOSMAC freq error: %s\n", TOSMAC_DEVICE);
		close(fd_tosmac);
		return -1;
	}	
	
	// set channel 11 to 26
	if(ioctl(fd_tosmac, TOSMAC_IOSETCHAN, CHANNEL) < 0)
	{
		fprintf (stderr, "IOCTL set TOSMAC channel error: %s\n", TOSMAC_DEVICE);
		close(fd_tosmac);
		return -1;
	}
	
/*
	// set local address
	if(ioctl(fd_tosmac, TOSMAC_IOSETADDR, srcAddr) < 0)
	{
		fprintf (stderr, "IOCTL set TOSMAC address error: %s\n", TOSMAC_DEVICE);
		close(fd_tosmac);
		return -1;
	}
	
	// enable auto ack
	ret = ioctl(fd_tosmac, TOSMAC_IOAUTOACK);
	if(ret < 0)
	{
		fprintf (stderr, "IOCTL enable TOSMAC AutoAck error: %s\n", TOSMAC_DEVICE);
		close(fd_tosmac);
		return -1;
	}
*/
	
	// disable auto ack
	ret = ioctl(fd_tosmac, TOSMAC_IODISAUTOACK);
	if(ret < 0)
	{
		fprintf (stderr, "IOCTL enable TOSMAC AutoAck error: %s\n", TOSMAC_DEVICE);
		close(fd_tosmac);
		return -1;
	}	
	
	//change max payload size (max value can be set is 116 bytes, change in "tosmacapi.h")
	if(ioctl(fd_tosmac, TOSMAC_IOSETMAXDATASIZE, MAX_PAYLOAD_SIZE) < 0)
	{
		fprintf (stderr, "IOCTL set TOSMAC max payload size error: %s\n", TOSMAC_DEVICE);
		close(fd_tosmac);
		return -1;
	}
	
	// set radio power level
	if(ioctl(fd_tosmac, TOSMAC_IOSETPOWER, pwr_level) < 0)
	{
		fprintf (stderr, "IOCTL set TOSMAC radio power level error: %s\n", TOSMAC_DEVICE);
		close(fd_tosmac);
		return -1;
	}
	
	// close device
	close(fd_tosmac);
	return 0;
}


/*
 * this function initializes a packet sturcture, it assigns a sequence # to the
 * packet (a random number)
 * inputs:
 *	destAddr: destination address
 *	type: packet type
 * outputs:
 *	returns 0 if successful
 */
int initPkt(packet *Pkt, int type)
{
	__u16 localAddr;	// local address
	
	localAddr = get_addr();
	memset(Pkt->data, 0x0, DATA_SIZE);	// set data payload to 0
	Pkt->dataSize = DATA_SIZE;
	Pkt->seq = rand();	// generate sequence number
	Pkt->type = type;	// set pkt type
	Pkt->srcAddr = localAddr;
	Pkt->rss = -127;	// set rss to min
	
	return 0;
}

/*
 * this functions sends a ACK package
 * inputs:
 * 	destAddr: destination address 
 * 	localAddr: local address
 * 	seq: sequence number of the package requesting ACK
 * 	fd_tosmac: currently opened radio device
 * outputs:
 * 	return 0 if correct ACK sent, return -1 for error
 * NOTICE radio device is not closed after this function returns
 */
int sendACK(__u16 destAddr, __u16 localAddr, unsigned char seq)
{
	int actSize = 0;	// actual size of data payload
	int i = 0;			// array index
	TOS_Msg pkt;		// package
	int fd_tosmac;		// radio device
	
	// Open radio device, read in non-blocking mode, write in blocking mode
	fd_tosmac = open(TOSMAC_DEVICE, O_RDWR|O_NONBLOCK);
	
	// initialize sending packet
	memset(&pkt, 0, sizeof(pkt));
	
	// initialize sending packet
	memset(&pkt, 0, sizeof(pkt));	
	pkt.length = MAX_PAYLOAD_SIZE;	// set data length of payload
	pkt.addr = destAddr;			// set dest address
	pkt.type = 2;					// set pkt type (1 for data, 2 for ack)
	pkt.group = seq;				// set group to be the seq #
	
	// NOTICE first 2 bytes of data section used to pass sender address
	// NOTICE bitwise operation used here, might be different on other platforms
	pkt.data[0] = localAddr >> 8;	// higher address
	pkt.data[1] = localAddr;		// lower address
	
	// send out ack
	if (write(fd_tosmac, &pkt, TOSH_DATA_LENGTH) >= 0) {
		printf("ACK sent... \n");
		close(fd_tosmac);	// close device
		return 0;
	}
	
	close(fd_tosmac);	// close device
	return -1;
}


/*
 * this functions receives a ACK package
 * inputs:
 * 	destAddr: destination address 
 * 	localAddr: local address
 * 	seq: sequence number of the package requesting ACK
 * 	fd_tosmac: currently opened radio device
 * outputs:
 * 	return 0 if correct ACK received within ACK_TIMEOUT seconds, -1 otherwise
 * NOTICE radio mutex is not included in this function. it is inside sendPkt 
 */
int recvACK(__u16 destAddr, __u16 localAddr, unsigned char seq)
{
	int i = 0;				// array index
	int fd_tosmac;			// radio device
	TOS_Msg pkt;			// package
	fd_set read_fds;		// set of file descriptors for select()
	int fdmax;				// maximum file descriptor value
	int ret;				// return value from select()
	struct timeval tv;		// time vector for timeout
	__u16 ackAddr;	// address of the received ack pkt
	__u16 tmp_16;	// 16 bit tmp var
	
	// Open radio device, read in non-blocking mode, write in blocking mode
	fd_tosmac = open(TOSMAC_DEVICE, O_RDWR|O_NONBLOCK);
	
	// set max timeout
	// NOTICE total waiting time for recvACK will be ACK_TIMEOUT
	tv.tv_sec = ACK_TIMEOUT;
	tv.tv_usec = RECV_TIMEOUT_U;
	
	// keep receving packets until timeout
	while (1) {
		// check for remaining time
		if (tv.tv_sec == 0)
			break;
		
		// Watch fd_tosmac to see when it has input.
		FD_ZERO(&read_fds);
		FD_SET(fd_tosmac, &read_fds);
		
		// maximum file descriptor value
		fdmax = fd_tosmac;
		
		// block until packet arrives
		ret = select(fdmax + 1, &read_fds, NULL, NULL, &tv);
		
		// check for select return val
		if (ret == -1) {
			printf("Error occored when waiting for ACK packet.\n");
			close(fd_tosmac);	// close device
			return -1;
		}
		
		// read packet when it arrives
		if(FD_ISSET(fd_tosmac, &read_fds)) {
			// initialize receiving packet
			memset(&pkt, 0, sizeof(pkt));
			
			// read received packet
			read(fd_tosmac, &pkt, TOSH_DATA_LENGTH);
			
			// show packet information
			if (network_debug) {
				printf("Length:%d ", pkt.length);
				printf("fcfhi:%x ", pkt.fcfhi);
				printf("fcflo:%x ", pkt.fcflo);
				printf("Seq#:%x ", pkt.dsn);
	//			printf("DestPAN:%x ", pkt.destpan);
				printf("DestAddr:%x ", pkt.addr);
				printf("SrcAddr:%x%x ", pkt.data[0], pkt.data[1]);
				printf("TypeID:%x ", (unsigned char) pkt.type);
				printf("GroupID:%x ", (unsigned char) pkt.group);
				printf("Strength:%d\n", pkt.strength);
	//			printf("LQI:%d ", pkt.lqi);
	//			printf("CRC:%d\n\n", pkt.crc);
			}
			
			// check if ack packet and dest
			if ((pkt.addr == localAddr) && (pkt.type == 2)) {
				// NOTICE bitwise operation used here
				ackAddr = pkt.data[0];		// get hi byte
				ackAddr = ackAddr << 8;	// shift to correct location
				tmp_16 = pkt.data[1];		// get lo byte
				ackAddr += tmp_16;			// combine 
				
				// check if ack is from the correct sender
				if(ackAddr == destAddr)	{	// if the ack is from correct addr
					if (pkt.group == seq) {	// sequence # matches
						printf("ACK received...\n");
						close(fd_tosmac);	// close device
						return 0;
					}
					else
						continue;
				}
				else
					continue;
			}
			else
				continue;
		}
		else {						// timeout
			printf("ACK timeout!\n");
			close(fd_tosmac);		// close device
			return -1;
		}
	}
	
	close(fd_tosmac);	// close device
	return -1;
}


/* 
 * this function sets up the imote2 radio and sends out a data packet
 * there are 3 types of data packets (data/ack/mac/beacon)
 * inputs:
 * 	destAddr: 16 bit address of dest, 0xFFFF is for broadcasting address
 * 	localAddr: 16 bit address of sender (src)
 * 	data: data to be sent, max size is MAX_PAYLOAD_SIZE - 2
 * 	dataSize: size of input data, in # of bytes
 * 	ACK_en: switch to turn on sending/receving ACK
 * outputs:
 * 	return 0 if send successful, otherwise return -1
 */
int sendPkt(__u16 destAddr, __u16 localAddr, unsigned char *data, int seq, int dataSize, int port, int ACK_en)
{
	int fd_tosmac;		// radio device
	int i = 0;			// index
	TOS_Msg pkt;		// tosmac package
	
	// initialize sending packet
	memset(&pkt, 0, sizeof(pkt));	
	pkt.length = MAX_PAYLOAD_SIZE;	// set data length of payload
//	pkt.fcfhi = 8;					// set FCF-hi to 0b00001000
//	pkt.fcflo = 0;					// set FCF-lo
//	pkt.dsn = 0;					// set initial sequence # (doesn't work)
//	pkt.destpan = 0;				// use destPAN for src addr (doesn't work)
	pkt.addr = destAddr;			// set dest address
	pkt.type = port;				// set pkt type (port number, 1 data, 2 ack)
	pkt.group = seq;				// set group #, reserved for ack
	
	// NOTICE first 2 bytes of data section used to pass sender address
	// NOTICE bitwise operation used here, might be different on other platforms
	pkt.data[0] = localAddr >> 8;	// higher address
	pkt.data[1] = localAddr;		// lower address
	
	// fill packet
	for (i = 2; i < dataSize + 2; i++) {
		pkt.data[i] = *data;
		data++;
	}
	
	// send out pkt
	i = 0;	// reset counter
	//pthread_mutex_lock(&radio_mutex);	// lock radio
	while(1) {
		// Open radio device, blocking
		fd_tosmac = open(TOSMAC_DEVICE, O_RDWR);
		
		//
		if (write(fd_tosmac, &pkt, TOSH_DATA_LENGTH) >= 0) {
			// close radio devie before receving ACK
			close(fd_tosmac);		// close device
			
			if (network_debug)
				printf("Packet being sent is: %s\n", (pkt.data + 2));
			
			// wait for a ACK
			if (ACK_en) {	// check ACK_en
				printf("Waiting for ACK...\n\n");
				// wait before receiving ACK
				usleep(WAIT_TIME);
				// 
				if (recvACK(destAddr, localAddr, pkt.group) == 0)
					break;	// ACK received
				else {
					// check max retries
					if (i++ > RETRIES) {	// update counter
						printf("Packet failed to send after retries.\n");
						close(fd_tosmac);
						//pthread_mutex_unlock(&radio_mutex);	// release radio
						return -1;
					}
					printf("Resending current package...\n");
					continue;				// resend packet
				}
			}
			else
				break;	// skip ACK
		}
		else {
			printf("Packet failed to send.\n");
			close(fd_tosmac);		// close device
			//pthread_mutex_unlock(&radio_mutex);	// release radio
			return -1;	// return -1 for transmission failure
		}
	}
	//pthread_mutex_unlock(&radio_mutex);	// release radio 
	return 0;
}


/*
* this function sets up the imote2 radio and receives data packets
* it is used for the handler function, it picks up any pakcets whose dest is
* the current node
* inputs:
* 	localAddr: 16 bit local address
* 	in_pkt: incoming packet struct
* 	ACK_en: switch to turn on sending/receving ACK
* outputs:
* 	return 0 if send successful, otherwise return -1
* 	return the RSS if received within ACK_TIMEOUT, -1 otherwise
*/
int h_recvPkt(__u16 localAddr, packet *in_pkt, int ACK_en)
{
	int fd_tosmac;			// radio device
	int actSize;			// actual size of data payload
	int i;					// array index
	TOS_Msg pkt;			// package
	fd_set read_fds;		// set of file descriptors for select()
	int fdmax;				// maximum file descriptor value
	int ret;				// return value from select()
	struct timeval tv;		// time vector for timeout
	__u16 destAddr;	// dest address for ack pkt
	__u16 tmp_16;	// 16 bit temp var
	unsigned char *tmp_data;//
	
	// Open radio device, read in non-blocking mode, write in blocking mode
	pthread_mutex_lock(&radio_mutex);	// lock radio
	fd_tosmac = open(TOSMAC_DEVICE, O_RDWR|O_NONBLOCK);
	
	// set max timeout
	// NOTICE total waiting time for recvACK will be ACK_TIMEOUT
	tv.tv_sec = RECV_TIMEOUT;
	tv.tv_usec = RECV_TIMEOUT_U;
	
	// keep receiving until gets packet or timeout
	while (1) {
		// check for remaining time
		if (tv.tv_sec == 0)
			break;
		
		// Watch fd_tosmac to see when it has input.
		FD_ZERO(&read_fds);
		FD_SET(fd_tosmac, &read_fds);
		
		// get maximum file descriptor value
		fdmax = fd_tosmac;
		
		// block until packet arrives
		ret = select(fdmax + 1, &read_fds, NULL, NULL, &tv);
		
		// check for select return val
		if (ret == -1) {
			printf("Error occored when waiting for ACK packet.\n");
			close(fd_tosmac);	// close device
			pthread_mutex_unlock(&radio_mutex);	// release radio
			return -1;
		}
		
		// read packet when it arrives
		if(FD_ISSET(fd_tosmac, &read_fds)) {
			// initialize receiving packet
			memset(&pkt, 0, sizeof(pkt));
			
			// read received packet
			read(fd_tosmac, &pkt, TOSH_DATA_LENGTH);
			
			// show packet information
			if (network_debug) {
				printf("Length:%d ", pkt.length);
				printf("fcfhi:%x ", pkt.fcfhi);
				printf("fcflo:%x ", pkt.fcflo);
				printf("Seq#:%x ", pkt.dsn);
				//printf("DestPAN:%x ", pkt.destpan);
				printf("DestAddr:%x ", pkt.addr);
				printf("SrcAddr:%x%x ", pkt.data[0], pkt.data[1]);
				printf("TypeID:%x ", (unsigned char) pkt.type);
				printf("GroupID:%x ", (unsigned char) pkt.group);
				printf("Strength:%d\n", pkt.strength);
				//printf("LQI:%d ", pkt.lqi);
				//printf("CRC:%d\n\n", pkt.crc);
			}
			
			// check packet dest
			if ((pkt.addr == localAddr) || (pkt.addr == 0xFFFF)) {
			//if (pkt.addr == localAddr) {
				// record packet type (port number)
				in_pkt->type = pkt.type;
				// get data from pkt
				tmp_data = in_pkt->data;
				for (i = 2; i < in_pkt->dataSize + 2; i++) {
					*tmp_data = pkt.data[i];
					tmp_data++;
				}
				in_pkt->seq = pkt.group;	// record seq #
				//*rss = pkt.strength;		// record rss
				printf("Packet received!\n");
				
				// get destAddr
				// NOTICE bitwise operation used here
				destAddr = pkt.data[0];		// get hi byte
				destAddr = destAddr << 8;	// shift to correct location
				tmp_16 = pkt.data[1];		// get lo byte
				destAddr += tmp_16;			// combine
				in_pkt->srcAddr = destAddr;	//
				in_pkt->rss = pkt.strength;	// get RSSI
				
				// clsoe radio before sending ACK
				close(fd_tosmac);			// close device
				
				// send a ACK
				// ACK not needed for routing packets
				if ((ACK_en) && (pkt.type != ROUTING_PKT_TYPE) && (pkt.type != TO_NODE_ROUTING_PKT_TYPE) && (pkt.type != ANCHOR_BROADCAST)) {
					for (i = 0; i < ACK_RESEND; i++) {	// send multiple ACKs if needed
						// wait before sending ACK, extra time added
						usleep(WAIT_TIME);
						usleep(WAIT_TIME);
						if (sendACK(destAddr, localAddr, pkt.group) < 0) {
							printf("Failed to send ack.\n\n");
							pthread_mutex_unlock(&radio_mutex);	// release radio
							return -1;
						}
					}
				}
				pthread_mutex_unlock(&radio_mutex);	// release radio
				return 0;
			}
			else
				continue;
		}
		else {
			printf("Receive timeout!\n\n");
			close(fd_tosmac);		// close device
			pthread_mutex_unlock(&radio_mutex);	// release radio
			return -1;
		}
	}
	
	close(fd_tosmac);		// close device
	pthread_mutex_unlock(&radio_mutex);	// release radio
	return -1;
}


/*
 * warpper for of send packet
 * inputs:
 *	destAddr: destination node address
 *	packet: packet struct to be sent
 *	ACK_en: ack enable
 * outputs:
 *	returns 0 if send successful, -1 otherwise
 */
int h_send(__u16 destAddr, packet *out_pkt, int ACK_en)
{
	// send packet
	pthread_mutex_lock(&radio_mutex);	// lock radio
	if (sendPkt(destAddr, out_pkt->srcAddr, out_pkt->data, out_pkt->seq, DATA_SIZE, out_pkt->type, ACK_en) < 0) {
		pthread_mutex_unlock(&radio_mutex);	// release radio
		printf("Packet failed to send to %x\n\n", destAddr);
		return -1;
	}
	pthread_mutex_unlock(&radio_mutex);	// release radio
	return 0;
}

