/*******************************************************************************
*	Multi-hop router. router.c
*	Last Modified: 21/May/2009
*
*	Copyright (C) 2009 Tony Wu mdzz@u.washington.edu
*
*	This program is free software: you can redistribute it and/or modify
*	it under the terms of the GNU General Public License as published by
*	the Free Software Foundation, either version 3 of the License, or
*	(at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	You should have received a copy of the GNU General Public License
*	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*******************************************************************************/

#include <string.h>
#include "router.h"
#include "socket.h"

// debug switch
#define routing_debug 0

// global vars
int numHops = 255;		// number of hops from base
Node *base_head = NULL;	// head pointer for base to node routing table (linked list)
int to_node_size = 0;	// to node routing table size
routingEntry toBaseTable[ROUTING_TABLE_SIZE];	// to base routing table


/*
 * this function initializes the routing table
 * return 0 if successful
 */
int initTable()
{
	int i;
	__u16 localAddr;
	baseRoutingEntry e;
	
	// initialize table to "0000 255 0"
	for (i = 0; i < ROUTING_TABLE_SIZE; i++) {
		toBaseTable[i].nextHop = 0;
		toBaseTable[i].numHops = 255;
		toBaseTable[i].lastHeard = 0;
	}
	
	// current node address
	localAddr = get_addr();
	
	printf("Node addrss is: %x\n", localAddr);
	if (check_base()) {
		numHops = 0;	// reset to 0 for base node
		printf("Current node is base node!\n\n");
	}
	else
		printf("\n");
	
	// initialize base routing table (linked list)
	e.size = 0;
	e.destAddr = 0x0;
	e.addrList = NULL;
	e.lastHeard = 0;
	base_head = append(base_head, e);
	
	return 0;
}


/*
 * this function sends routing packet to all other nodes.
 * Content of the routing packet is read from "router.conf" file.
 * Routing packets have type ID 9.
 * output:
 *	returns 0 if send successful, -1 otherwise 
 */
int send_to_base_rtPkt()
{
	int i;
	packet out_pkt;		// out going packet
	__u16 localAddr;
	unsigned char data[ROUTING_DATA_SIZE];	//
	
	// get localAddr
	localAddr = get_addr();
	
	// initialize packet
	initPkt(&out_pkt, ROUTING_PKT_TYPE);
	
	// build data payload
	memcpy(out_pkt.data, &localAddr, 2);	//
	out_pkt.data[2] = numHops;		// notice that this number is only 8 bits (255)
	if (routing_debug) {
		printf("Current node address is: %x\n", out_pkt.data[0]);
		printf("Num of hops is: %d\n", out_pkt.data[2]);
	}
	// read routing table into routing packet
	for (i = 0; i < ROUTING_TABLE_SIZE; i++) {
		memcpy(out_pkt.data + 2 * i + 3, &(toBaseTable[i].nextHop), 2);
		if (routing_debug)
			printf("Can hear: %x\n", out_pkt.data[2 * i + 3]);
	}
	
	// send back-to-base routing packet
	if (h_send(BROADCAST_ADDR, &out_pkt, 0) < 0) {
		printf("Failed sending to-base routing packet!\n\n");
		return -1;
	}
	else
		printf("Back to base routing packets sent.\n\n");
	
	return 0;
}


/*
 * this function sends routing packets back to basenode. Used for base to node
 * routing.
 */
int send_to_node_rtPkt()
{
	int i;
	packet out_pkt;		// out going packet
	__u16 destNode, localAddr;
	
	// get localAddr
	localAddr = get_addr();
	
	// read local to base routing table
	for (i = 0; i < ROUTING_TABLE_SIZE; i++){
		// first non-zero entry is always the shortest
		destNode = toBaseTable[i].nextHop;
		if (destNode != 0x0)
			break;
	}
	// check if path exists
	if (destNode == 0x0) {
		printf("No path to base.\n\n");
		return -1;
	}
	
	// initialize packet
	initPkt(&out_pkt, TO_NODE_ROUTING_PKT_TYPE);
	
	// build data payload
	// NOTICE bitwise operation used here
	out_pkt.data[0] = 1;	// # of nodes contained in the packet, max 8
	out_pkt.data[1] = localAddr >> 8;	// high bits of address
	out_pkt.data[2] = localAddr;		// low bits of address
	
	// send routing packet
	if (sendtoBase(&out_pkt, 0) < 0) {
		printf("Failed sending routing packet!\n\n");
		return -1;
	}
	else
		printf("Base to nodes routing packet sent.\n\n");
	
	return 0;
}


/*
 * warpper for sending routing beacons every MIN_UPDATE_INTERVAL seconds
 */
void *routingBeacon()
{
	int i;
	extern sendCommand;		// send command flag
	while (1) {
		if (!sendCommand) {	// stop beaconing routing info when sending commands
			// send route to base packet
			for (i = 0; i < TO_BASE_ROUTING_RESEND; i++) {
				if (send_to_base_rtPkt() < 0)
					printf("Error sending routing packets.\n");
				// wait before next send
				//usleep(WAIT_TIME);
			}
			usleep(MIN_UPDATE_INTERVAL);	//
			// send route to node packet
			if (!check_base()) {
				for (i = 0; i < TO_NODE_ROUTING_RESEND; i++) {
					if (send_to_node_rtPkt() < 0)
						printf("Error sending routing packets.\n");
					// wait before next send
					//usleep(WAIT_TIME);
				}
			}
			usleep(MIN_UPDATE_INTERVAL);	//
		}
	}
	pthread_exit(NULL);		// exit thread
}


/*
 * this function updates the to base routing table according to the received
 * routing msg
 * inputs:
 *	cur_node: current node properties (addr, # of hops from base)
 *	src_node: source node properties
 *	canHear: list of nodes that can be heard from src_node
 * outputs:
 *	returns 0 if update successful, -1 otherwise
 */
int updateTable(nodeProperty src_node, __u16 *canHear)
{
	int neighbor_can_hear = 0;	// flag set if neighbor can hear current node
	int i, j;
	__u16 tmp_nextHop, localAddr;
	__u8 tmp_numHops;
	int tmp_lastHeard;
	
	// get localAddr
	localAddr = get_addr();
	
	// if sender is basenode
	if (src_node.numHops == 0)
		numHops = 1;	// then current node is 1 hop away
	
	// check if neighbor can hear current node
	for (i = 0; i < ROUTING_TABLE_SIZE; i++) {
		if (localAddr == canHear[i]) {
			neighbor_can_hear = 1;
			break;
		}
	}
	
	// add new neighbor to routing table
	if (neighbor_can_hear) {
		for (i = 0; i < ROUTING_TABLE_SIZE; i++) {
			// replace the larger entry
			if ((toBaseTable[i].numHops >= src_node.numHops) || (toBaseTable[i].nextHop == src_node.addr)) {
				toBaseTable[i].nextHop = src_node.addr;
				toBaseTable[i].numHops = src_node.numHops;
				toBaseTable[i].lastHeard = 0;
				// update current node numHops
				if (numHops >=  src_node.numHops + 1)
					numHops = src_node.numHops + 1;
				break;
			}
		}
	}
	// neighbor CANT hear current node, add neighbor to routing table
	else {
		for (i = 0; i < ROUTING_TABLE_SIZE; i++) {
			// replace default entry
			if ((toBaseTable[i].numHops == 255) || (toBaseTable[i].nextHop == src_node.addr)) {
				toBaseTable[i].nextHop = src_node.addr;
				toBaseTable[i].numHops = src_node.numHops;
				toBaseTable[i].lastHeard = 0;
				// update current node numHops
				if (numHops >=  src_node.numHops + 1)
					numHops = src_node.numHops + 1;
				break;
			}
		}
	}
	
	// update lastHeard
	for (i = 0; i < ROUTING_TABLE_SIZE; i++) {
		toBaseTable[i].lastHeard++;
		if (toBaseTable[i].lastHeard > MAX_STAY) {
			toBaseTable[i].nextHop = 0;	// reset to initial
			toBaseTable[i].numHops = 255;	// reset to max
			toBaseTable[i].lastHeard = 0;
		}
	}
	
	// sort routing table (bubble sort)
	for (i = 0; i < ROUTING_TABLE_SIZE; i++) {
		for (j = i + 1; j < ROUTING_TABLE_SIZE; j++) {
			if (toBaseTable[i].numHops > toBaseTable[j].numHops) {
				// swap values
				tmp_nextHop = toBaseTable[i].nextHop;
				tmp_numHops = toBaseTable[i].numHops;
				tmp_lastHeard = toBaseTable[i].lastHeard;
				toBaseTable[i].nextHop = toBaseTable[j].nextHop;
				toBaseTable[i].numHops = toBaseTable[j].numHops;
				toBaseTable[i].lastHeard = toBaseTable[j].lastHeard;
				toBaseTable[j].nextHop = tmp_nextHop;
				toBaseTable[j].numHops = tmp_numHops;
				toBaseTable[j].lastHeard = tmp_lastHeard;
			}
		}
	}
	
	return 0;
}


/*
 * this function sends packets back to the base node via routing
 * inputs:
 *	out_pkt: pointer for out going packet
 *	ACK_en: switch for truning on/off ACK
 * output:
 *	return 0 if successful, -1 otherwise
 */
int sendtoBase(packet *out_pkt, int ACK_en)
{
	int i;
	__u16 destNode;	// destination Node address
	
	// read local to base routing table
	for (i = 0; i < ROUTING_TABLE_SIZE; i++){
		// first non-zero entry is always the shortest
		destNode = toBaseTable[i].nextHop;
		if (destNode != 0x0)
			break;
	}
	// check if path exists
	if (destNode == 0x0) {
		printf("No path to base.\n\n");
		return -1;
	}
	
	// send pkt
	printf("Sending packet back to base through: %x\n", destNode);
	if (h_send(destNode, out_pkt, ACK_en) < 0)
		return -1;
	
	return 0;
}


/*
 * this function sends command data to destination nodes via routing
 * the actuall command data is embedded into the last 9 bytes of the data
 * payload, # of hops left is recorded in the first byte of the data payload.
 * everytime the packet passes through a node, the number is reduced by 1.
 */
int sendtoNode(__u16 destAddr, packet *out_pkt, int ACK_en)
{
	int i;
	packet tmp_pkt;				//
	baseRoutingEntry *target;	// base to node routing entry
	unsigned char data[ROUTING_DATA_SIZE];
	__u16 nextHop;
	
	// initialize packet
	initPkt(&tmp_pkt, CMD_PKT_TYPE);
	
	// move cmd data to the last 12 bytes
	memcpy(tmp_pkt.data + 14, out_pkt->data, 12);
	
	// fill in routing section
	target = return_entry(base_head, destAddr);
	if (target == NULL) {
		printf("No path to node %x.\n\n", destAddr);
		return -1;
	}
	else {
		printf("Found path to node %x.\n", target->destAddr);
		printf("%d hops needed.\n", target->size - 1);
	}
	tmp_pkt.data[0] = target->size;
	tmp_pkt.data[1] = target->destAddr >> 8;	// fill dest addr
	tmp_pkt.data[2] = target->destAddr;
	
	for (i = 0; i <= target->size - 2; i++) {	// fill the routing list
		tmp_pkt.data[3 + 2 * i] = target->addrList[i] >> 8;
		tmp_pkt.data[4 + 2 * i] = target->addrList[i];
	}
	if ((target->size - 1) == 0)	// 1 hop away
		nextHop = target->destAddr;
	else
		nextHop = target->addrList[target->size - 2];
	
	// send pkt
	printf("Sending packet to %x through: %x\n\n", destAddr, nextHop);
	if (h_send(nextHop, &tmp_pkt, ACK_en) < 0)
		return -1;
	
	// print out each byte of the cmd packet
	if (routing_debug) {
		printf("\n");
		for (i = 0; i < 26; i++) {
			printf("To node command packet content is: %x\n", tmp_pkt.data[i]);
		}
		printf("\n");
	}
	return 0;
}


/*
 * this function handles the incoming to base routing packets
 */
int recv_rtPkt(packet *in_pkt)
{
	unsigned char *tmp;
	__u16 tmp_16;
	int i;
	__u16 canHear[ROUTING_TABLE_SIZE];	// table of reachable nodes
	nodeProperty src_node;	// msg node
	
	// get srcAddr, srcHop
	memcpy(&(src_node.addr), in_pkt->data, 2);	//
	src_node.numHops = in_pkt->data[2];
	if (routing_debug) {
		printf("\nRouting pakcets from: %x, %d hops away from base node.\n", src_node.addr, src_node.numHops);
	}
	
	// get all neighbor address
	tmp = in_pkt->data + 3;
	for (i = 0; i < ROUTING_TABLE_SIZE; i++) {
		memcpy(&(canHear[i]), tmp, 2);	//
		tmp += 2;
		if (routing_debug)
			printf("Neighbor address are: %x\n", canHear[i]);
	}
	// update routing table
	if (updateTable(src_node, canHear) < 0) {
		printf("Error updating routing table! Program terminated!\n");
		return -1;
	}
	
	return 0;
}


/*
 * this function forwards routing packets to base
 * used for base to node routing
 */
int forward_rt_toBase(packet *out_pkt)
{
	int i;
	__u16 localAddr;
	pthread_t routing_thread;
	
	// get localAddr
	localAddr = get_addr();
	
	if (check_base()) {	// process packet if currently at base node
		printf("Routing packets have reached base node!\n\n");
		base_recv_rtPkt(out_pkt);
	}
	else {	// forward msg to next hop
		out_pkt->srcAddr = localAddr;
		out_pkt->data[0]++;		// update # of nodes in pkt
		i = (out_pkt->data[0] - 1) * 2;
		if (i > (MAX_NUM_NODES_IN_RT * 2)) {	// stop if maximum number excedded
			printf("Maximum number of nodes reached!\n");
			return -1;
		}
		// NOTICE bitwise operation
		// attach current node address to the END of the list
		out_pkt->data[i + 1] = localAddr >> 8;
		out_pkt->data[i + 2] = localAddr;
		if (sendtoBase(out_pkt, 0) < 0)
			return -1;
	}
	return 0;
}


/*
 * this function receives base-to-node routing packets at the base node
 * routing information is stored locally in a routing table (linked list)
 * input:
 * 	in_pkt: received routing packet
 * output:
 * 	return 0 if update success
 */
int base_recv_rtPkt(packet *in_pkt)
{
	baseRoutingEntry routing_node;
	Node *p;
	int i;
	
	routing_node.lastHeard = 0;	//
	routing_node.size = in_pkt->data[0];	// get # of hops 
	// obtain routing list
	routing_node.addrList = (__u16 *)malloc((routing_node.size - 1) * 4);
	for (i = 0; i < routing_node.size - 1; i++) {
		routing_node.addrList[i] = in_pkt->data[3 + 2 * i];
		routing_node.addrList[i] = routing_node.addrList[i] << 8;
		routing_node.addrList[i] += in_pkt->data[4 + 2 * i];
	}
	// NOTICE bitwise operation
	// get destination node (1st element in routing list)
	routing_node.destAddr = in_pkt->data[1];
	routing_node.destAddr = routing_node.destAddr << 8;
	routing_node.destAddr += in_pkt->data[2];
	
	printf("Received base routing packet from: %x\n", routing_node.destAddr);
	
	// add/update base rouitng table
	if ((routing_node.size < MAX_NUM_NODES_IN_RT) && (routing_node.size != 0))
		append(base_head, routing_node);
	else
		printf("Maximum number of hops exceeded.\n");
	
	// check routing table and remove old entries
	check_node(base_head);
	
	// print current routing table
	if (routing_debug) {
		printf("Base routing table is:\n");
		for (p = base_head; p; p = p->next) {
			printf("Address: %x, %d hops away. ", p->e.destAddr, p->e.size);
			for (i = 0; i < p->e.size - 1; i++)
				printf("%x ", p->e.addrList[i]);
			printf("\n");
		}
		printf("\n");
	}
	
	return 0;
}


/*
 * this function addes a node to the linked list.
 * inputs:
 * 	h: head of list
 * outputs:
 *	p: pointer to node added
 */
Node *append(Node *h, baseRoutingEntry e)
{
	Node *prev = h;
	Node *p = h;
	int update = 0;	// flag for updating internal node
	
	if (h)
		p = p->next;
	
	while (p) {
		prev = p;
		if (p->e.destAddr == e.destAddr) {
			update = 1;	// found existing node
			break;
		}
		p = p->next;
	}

	if (!update) {	// add new node to the end of the list
		p = (Node *)malloc(sizeof(Node));
		p->e = e;
		p->next = NULL;
		if (prev)	//
			prev->next = p;
	}
	else {	// update existing node
		p->e.size = e.size;
		p->e.addrList = e.addrList;
		p->e.lastHeard = 0;
	}
	
	return p;
}


/*
 * this function walks through the routing table (linked list) returns the
 * routing table entry for the given address.
 */
baseRoutingEntry *return_entry(Node* h, __u16 addr)
{
	baseRoutingEntry target;
	Node *p = h;
	
	if (h)
		p = p->next;	// skip first element
	while (p) {
		if (p->e.destAddr == addr)
			return &p->e;
		p = p->next;
	}
	
	return NULL;
}


/*
 * this function deletes node from the linked list
 */
int delete_node(Node *prev, Node *cur)
{
	prev->next = cur->next;	// link to next
	free(cur);	// free memory
	return 0;
}


/*
 * this function checks each entry in the to-node routing table and deletes old
 * entries
 */
int check_node(Node *h)
{
	Node *prev = h;
	Node *cur = h;
	Node *tmp = NULL;
	int table_size = 0;
	
	if (cur)
		cur = cur->next;	// skip first entry
	
	while (cur) {
		cur->e.lastHeard++;	// update lastHeard
		tmp = cur;
		cur = cur->next;	// move to next
		if (tmp->e.lastHeard > MAX_STAY_TO_NODE) {
			if (routing_debug)
				printf("Removing %x from to-node routing table!\n", tmp->e.destAddr);
			delete_node(prev, tmp);	// delete old node
		}
		else
			prev = tmp;		//
		table_size++;	// count table size
	}
	
	// update to node routing table size
	to_node_size = table_size;
	return 0;
}


/*
 * helper thread to display routing table periodcally
 */
void *display_routing_table()
{
	int i, j;
	__u16 routingTable[127];	// at most 127 entries
	Node *p;
	
	while (1) {
		// show node to base routing table
		if (!check_base()) {
			printf("\nNode-to-base routing table is:\n");
			for (i = 0; i < ROUTING_TABLE_SIZE; i++)
				printf("%x\t%d\t%d\n", toBaseTable[i].nextHop, toBaseTable[i].numHops, toBaseTable[i].lastHeard);
			printf("\n");
		}
		else {
			// show to-node routing table
			printf("Base-to-nodes routing table is:\n");
			//
			j = 0;
			for (p = base_head->next; p; p = p->next) {
				printf("Address: %x, %d hops away, last heard: %d. ", p->e.destAddr, p->e.size, p->e.lastHeard);
				for (i = 0; i < p->e.size - 1; i++)
					printf("%x ", p->e.addrList[i]);
				printf("\n");
				// build routingTable
				routingTable[j] = p->e.destAddr;
				j++;
				if (j >= 127) {
					printf("\nTo-node routing table to large!\n\n");
					exit(-3);
				}
			}
			printf("\n\n");
			// send to node routing table back to GUI
			routingSocket(routingTable, to_node_size);
		}
		sleep(TABLE_DISPLAY);	// display intival
	}
}
